#!/usr/bin/env python
## @file
#
# Python Frontend for MobCom client
#
# (C) 2008-06-04 Georg Lukas <glukas>
#
# This is the official frontend module used for the text mode interface
# communication.
#
# This code is published under the GNU General Public Licence v2.

import pygtk
pygtk.require('2.0')
import gtk
import gobject

import sys
import socket
import time

## Allow absolute path names for the SEND command
SEND_ABSOLUTE_PATH = False

## Raise windows on action (message / incoming file)
RAISE_WINDOWS = False

## Log color for incoming messages
COLOR_IN="#ff0000"
## Log color for sent messages
COLOR_OUT="#0000ff"

## backend connection TCP address
BACKEND=('localhost', 5507)

## broadcast MAC address
BC = "ff:ff:ff:ff:ff:ff"

## This class implements a chat window for interactive private or broadcast chatting.
class Chat:
	## Callback, when the peer changes their name
	def update_name(self, name):
		oldname = self.name
		self.name = name
		self.win.set_title("Chat: %s <%s>" % (self.name, self.mac))
		if oldname != self.name:
			self.log("*** " + oldname + " is now known as " + self.name + "\n")

	## Constructor for the Chat window
	# @param fe Frontend responsible for this Chat
	# @param peer tuple of (MAC, name)
	def __init__(self, fe, peer):
		## Frontend pointer
		self.fe = fe
		## MAC address of the peer (as string), name of the peer
		self.mac, self.name = peer
		## GTK window associated with this Chat
		self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.update_name(self.name)
		self.win.set_default_size(600, 400)
		self.win.connect("destroy", self.cb_destroy)

		# chat elements
		self.text = gtk.TextBuffer()
		self.text.create_tag("in", foreground=COLOR_IN)
		self.text.create_tag("out", foreground=COLOR_OUT)
		self.textview = gtk.TextView(self.text)
		self.textview.set_editable(False)

		hboxin = gtk.HBox(True, 0)
		hboxin.set_homogeneous(False)
		self.input = gtk.Entry()
		self.input.set_activates_default(True)

		self.enter = gtk.Button("Ok")
		self.enter.connect("clicked", self.cb_enter)
		self.enter.set_flags(gtk.CAN_DEFAULT)

		hboxin.pack_start(self.input, True, True)
		hboxin.pack_end(self.enter, False, True)


		vbox = gtk.VBox(False, 0)
		sw = gtk.ScrolledWindow()
		sw.add(self.textview)
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
		vbox.pack_start(sw, True, True)
		vbox.pack_start(hboxin, False, True)

		self.win.add(vbox)
		self.win.show_all()
		self.enter.grab_default()
		self.input.grab_focus()

	## Callback for activating and deactivating buttons
	def buttonstate(self, is_conn):
		self.enter.set_sensitive(is_conn)

	## Write a formatted line of text into the scrolling text window
	# @param text line of text
	# @param tag color formatting tag (possible values are "in" and "out")
	def log(self, text, tag=None):
		self.text.insert(self.text.get_end_iter(), time.strftime("%H:%M:%S "))
		ei = self.text.get_end_iter()
		if tag == None:
			self.text.insert(ei, text)
		else:
			self.text.insert_with_tags_by_name(ei, text, tag)
		self.textview.scroll_mark_onscreen(self.text.get_insert())

	## Display a message received from a user
	# @param nick name of the user
	# @param text message
	def got(self, nick, text):
		self.log("<%s> %s\n" % (nick, text), "in")

	## Callback for closing the window
	def cb_destroy(self, widget, data=None):
		del self.win
		self.fe.closed_chat(self.mac)
		return False

	## Callback for pressing <Enter>
	def cb_enter(self, widget, data=None):
		msg = self.input.get_text()
		if len(msg) == 0:
			return
		self.input.set_text("")
		self.fe.send("MSG %s %s\n" % (self.mac, msg))
		self.log("<%s> %s\n" % (self.fe.name, msg), "out")

STAT_OFFER = "Offered..."

## A window for handling file downloads.
class FileDownload:
	## Callback, when the peer changes their name
	def update_name(self, name):
		self.name = name
		self.win.set_title("File Download: %s <%s>" % (self.name, self.mac))

	def init_filelist(self):
		self.filestore = gtk.TreeStore(str, int, str)

		self.filetree = gtk.TreeView(self.filestore)
		cnames = ('File name', 'Size', 'Status')
		for col in (0, 1, 2):
			tvc = gtk.TreeViewColumn(cnames[col])
			crt = gtk.CellRendererText()
			tvc.pack_start(crt, True)
			tvc.add_attribute(crt, 'text', col)
			tvc.set_sort_column_id(col)

			self.filetree.append_column(tvc)

		self.filetree.set_search_column(0)
		self.filetree.set_reorderable(True)
		#self.filetree.set_activates_default(True)

	def add_offer(self, filename, size):
		#print [filename, int(size), STAT_OFFER]
		self.filestore.append(None, [filename, int(size), STAT_OFFER])

	def __init__(self, fe, peer):
		self.fe = fe
		self.mac = peer[0]
		self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.update_name(peer[1])
		self.win.set_default_size(630, 200)
		self.win.connect("destroy", self.cb_destroy)

		self.init_filelist()

		hbox = gtk.HBox(True, 0)
		hbox.set_homogeneous(False)

		self.start = gtk.Button("_Start")
		self.start.connect("clicked", self.cb_start)
		self.start.set_flags(gtk.CAN_DEFAULT)

		self.filetree.connect("cursor-changed", self.cb_select)

		self.ignore = gtk.Button("_Ignore")
		self.ignore.connect("clicked", self.cb_destroy)

		hbox.pack_start(self.start, True, True)
		hbox.pack_start(self.ignore, True, True)


		vbox = gtk.VBox(False, 0)
		sw = gtk.ScrolledWindow()
		sw.add(self.filetree)
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
		vbox.pack_start(sw, True, True)
		vbox.pack_start(hbox, False, True)

		self.win.add(vbox)
		self.win.show_all()
		self.start.grab_default()
		self.filetree.grab_focus()
		self.current = None
		self.buttonstate(True)

	## Get the filename of the selected filestore item
	# @return filename of the selected item or None
	def get_sel_file(self, update_current=False):
		(ts, tsi) = self.filetree.get_selection().get_selected()
		# nothing selected, return None
		if tsi is None:
			return None
		# current should be set to the selected one, if not already set
		if update_current:
			if self.current is not None:
				return None
			self.current = tsi
		return ts.get(tsi, 0)[0]

	def set_status(self, status):
		if self.current:
			f = self.filestore[self.current]
			f[2] = status

	def abort(self, status):
		self.set_status(status)
		self.current = None

	def done(self, status):
		self.set_status(status)
		self.current = None

	## Callback for activating and deactivating buttons
	def buttonstate(self, is_conn=None):
		if is_conn is not None:
			self.btnstate = is_conn
		tsi, ts = self.filetree.get_selection().get_selected()
		if ts == None:
			self.start.set_sensitive(False)
			self.ignore.set_sensitive(False)
		else:
			self.start.set_sensitive(self.btnstate)
			self.ignore.set_sensitive(True)

	def cb_start(self, widget, data=None):
		fn = self.get_sel_file(True)
		if fn:
			self.fe.send("GET %s %s\n" % (self.mac, fn))
			#self.current = fn
			self.set_status("Downloading...")
			self.buttonstate()
		return False

	## Callback for closing the window
	def cb_destroy(self, widget, data=None):
		del self.win
		self.fe.closed_file(self.mac)
		return False

	def cb_select(self, widget, data=None):
		self.buttonstate()
		return False


COL_PEER = 0
COL_MAC  = 1
COL_AGE  = 2
COL_NAMES = ('Peer', 'MAC', 'Age')

## The actual Frontend implementation, responsible for all communication and widgets
class Frontend:
	## Create a TreeStore containing all known peers
	def init_peertree(self):
		self.peerstore = gtk.TreeStore(str, str, int)
		self.peerstore.append(None, ["(broadcast)", BC, 0])

		self.peertree = gtk.TreeView(self.peerstore)
		for col in (COL_PEER, COL_MAC, COL_AGE):
			tvc = gtk.TreeViewColumn(COL_NAMES[col])
			crt = gtk.CellRendererText()
			tvc.pack_start(crt, True)
			tvc.add_attribute(crt, 'text', col)
			tvc.set_sort_column_id(col)

			self.peertree.append_column(tvc)

		self.peertree.set_search_column(COL_PEER)
		self.peertree.set_reorderable(True)

	## Create a main Window
	def init_gui(self):
		self.win = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.win.set_title("MobKom GUI Frontend")
		self.win.set_default_size(720, 500)
		self.win.connect("destroy", self.cb_destroy)
		bnames = ('Connect', 'Chat', 'Send', 'Quit')
		self.btn = {}
		for bn in bnames:
			self.btn[bn] = gtk.Button("_" + bn)
		self.btn["Quit"].connect_object("clicked", gtk.Widget.destroy,
			self.win)
		self.btn["Connect"].connect("clicked", self.cb_connect)
		self.btn["Chat"].connect("clicked", self.cb_chat)
		self.btn["Send"].connect("clicked", self.cb_send)

		self.init_peertree()

		vpaned = gtk.VPaned()
		sw = gtk.ScrolledWindow()
		sw.add(self.peertree)
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

		vpaned.pack1(sw, True, True)

		self.text = gtk.TextBuffer()
		self.text.create_tag("in", foreground=COLOR_IN)
		self.text.create_tag("out", foreground=COLOR_OUT)
		self.textview = gtk.TextView(self.text)
		self.textview.set_editable(False)
		sw = gtk.ScrolledWindow()
		sw.add(self.textview)
		sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
		vpaned.pack2(sw, True, True)

		vbox = gtk.VBox(False, 0)
		vbox.pack_start(vpaned, True, True)

		hboxbtn = gtk.HBox(True, 0)
		for bn in bnames:
			hboxbtn.pack_start(self.btn[bn], True, True)
		vbox.pack_start(hboxbtn, False, True)

		# show everything in the window
		self.win.add(vbox)
		self.win.show_all()

	## Constructor of the Frontend
	def __init__(self):
		self.init_gui()
		self.name = "ich habe START vergessen"
		self.mac = BC
		self.be = None
		self.be_id = None
		self.chats = {}
		self.files = {}
		self.toggle_connect()
		gobject.timeout_add(1000, self.update_age)

	## Write a formatted line of text into the scrolling text window
	# @param text line of text
	# @param tag color formatting tag (possible values are "in" and "out")
	def log(self, text, tag=None):
		ei = self.text.get_end_iter()
		if tag == None:
			self.text.insert(ei, text)
		else:
			self.text.insert_with_tags_by_name(ei, text, tag)
		self.textview.scroll_mark_onscreen(self.text.get_insert())

	## Send a command to the backend (main application) and log it
	def send(self, request):
		self.be.send(request)
		self.log(">>> " + request, "out")

	## Periodic callback to increase the age of all peers
	def update_age(self):
		for node in self.peerstore:
			if node[COL_MAC] != BC: node[COL_AGE] += 1
		return True

	def join_peer(self, mac, nick):
		self.peerstore.append(None, [nick, mac, 0])

	def part_peer(self, mac, nick):
		#print "part:", mac, nick
		iter = self.peerstore.get_iter_first()
		while iter:
			#print self.peerstore.get(iter, COL_PEER)[0], nick
			if self.peerstore.get(iter, COL_PEER)[0] == nick:
				self.peerstore.remove(iter)
				return
			iter = self.peerstore.iter_next(iter)

	## Callback when a HELLO is received
	# @param mac MAC address of the peer
	# @param nick updated name
	def update_peer(self, oldnick, newnick):
		mac = self.peer_mac(oldnick)
		if mac in self.chats:
			self.chats[mac].update_name(newnick)
		if mac in self.files:
			self.files[mac].update_name(newnick)
		for node in self.peerstore:
			if node[COL_MAC] == mac:
				node[COL_PEER] = newnick
				node[COL_AGE] = 0
				return
		#self.peerstore.append(None, [nick, mac, 0])

	## Find the nick of a peer
	# @param mac MAC address to look for
	# @return name if MAC was found or mac
	def peer(self, mac):
		for node in self.peerstore:
			if node[1] == mac:
				return node[0]
		return mac

	def peer_mac(self, nick):
		for node in self.peerstore:
			if node[0] == nick:
				return node[1]
		return None

	## Get a [MAC, nick] tuple of the selected peerstore item
	# @return tuple of [MAC, nick] for selected item or broadcast if none selected
	def get_sel_peer(self):
		(ts, tsi) = self.peertree.get_selection().get_selected()
		if tsi is None:
			return [BC, "(broadcast)"]
		return ts.get(tsi, 1, 0)

	## Parse a new incoming message from the backend
	# This function parses a received message from the main application.
	# It looks for START, HELLO, MSG and FOFFER messages and generates the
	# right GUI events.
	#
	# @param msg ASCII message from backend (can contain multiple lines)
	def parse_be(self, msg):
		for l in msg.splitlines():
			# Hack for students sending latin-1 umlauts
			try:
				lutf8 = l.decode('utf-8')
			except UnicodeDecodeError:
				lutf8 = l.decode('latin-1')

			self.log("<<< %s\n" % lutf8, "in")
			args = lutf8.split(" ", 5)
			if len(args) > 5:
				(type, dest, nick, seq, sizeparam, text) = args
			else:
				(type, dest, nick) = args
				seq = sizeparam = text = None
			type = type.upper()
			src_mac = self.peer_mac(nick)

			if type == "START":
				self.mac = dest
				self.name = nick
				self.log("I am %s (%s)\n" % (self.name, self.mac))
			elif type == "JOIN":
				self.join_peer(dest, nick);
			elif type == "PART":
				self.part_peer(dest, nick);
			elif type == "NAME":
				self.update_peer(nick, dest);
			elif type == "MSG":
				if (dest.lower() == BC):
					self.chat(BC).got(nick, text)
				elif (dest == self.mac):
					self.chat(src_mac).got(nick, text)
				else:
					self.chat(BC).got(nick + " - " + self.peer(dest), text)
			elif type == "FOFFER":
				self.file(src_mac).add_offer(text, sizeparam)
			elif type == "FABORT":
				self.file(src_mac).abort(text)
			elif type == "FRECV":
				self.file(src_mac, raise_=False).set_status(text)
			elif type == "FDONE":
				self.file(src_mac).done(text)


	## Callback when the main window is closed
	def cb_destroy(self, widget, data=None):
		if self.be != None:
			self.be.close()
		gtk.main_quit()

	## Callback for activating and deactivating buttons
	def buttonstate(self, is_conn):
		names = { False: "_Connect", True: "_Disconnect" }
		self.btn['Connect'].set_label(names[is_conn])
		self.btn['Send'].set_sensitive(is_conn);
		#self.btn['Send'].set_sensitive(False);
		self.btn['Chat'].set_sensitive(is_conn);
		for ch in self.chats.itervalues():
			ch.buttonstate(is_conn)
		for fd in self.files.itervalues():
			fd.buttonstate(is_conn)

	## Function for opening and closing the backend socket
	# @param msg message to be displayed as reason
	def toggle_connect(self, msg=None):
		if self.be == None:
			self.log("Connecting to %s:%i... " % BACKEND)
			try:
				self.be = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
				self.be.connect(BACKEND)
				self.be_id = gobject.io_add_watch(self.be, gobject.IO_IN, self.cb_input)
				self.log("Ok.\n")
				self.buttonstate(True);
			except socket.error, err:
				self.log("Error: %s\n" % err[1])
				self.be = None
				if self.be_id != None:
					gobject.source_remove(self.be_id)
					self.be_id = None
		else:
			if self.be_id != None:
				gobject.source_remove(self.be_id)
				self.be_id = None
			self.be.close()
			self.log("Disconnected: %s\n" % msg)
			self.be = None
			self.buttonstate(False);

	## Callback for [Connect] button
	def cb_connect(self, widget, data=None):
		self.toggle_connect("User request")

	## Callback for the backend data socket
	def cb_input(self, fd, cond):
		try:
			data = fd.recv(128*1024)
			if data == '':
				be = None
				return False
			self.parse_be(data)
			return True
		except socket.error, err:
			self.toggle_connect(err[1])
			return False

	## Find the Chat window for a given peer, create it if needed.
	# @param mac MAC address of the peer
	# @param name optional name for a newly created window
	# @return reference to a Chat instance
	def chat(self, mac, name=None):
		if mac not in self.chats:
			if name is None:
				name = self.peer(mac)
			self.chats[mac] = Chat(self, (mac, name))
		if RAISE_WINDOWS:
			self.chats[mac].win.window.raise_()
		return self.chats[mac]

	## Delete a Chat from the lookup table
	# @param mac MAC address to be deleted
	def closed_chat(self, mac):
		del self.chats[mac]

	## Find and return a FileDownload window. Create if needed.
	def file(self, mac, name=None, raise_=True):
		if mac not in self.files:
			if name is None:
				name = self.peer(mac)
			self.files[mac] = FileDownload(self, (mac, name))
		if RAISE_WINDOWS and raise_:
			self.files[mac].win.window.raise_()
		return self.files[mac]

	def closed_file(self, mac):
		del self.files[mac]

	## Callback for the [Chat] button
	def cb_chat(self, widget, data=None):
		(mac, name) = self.get_sel_peer();
		self.chat(mac, name)
		#self.send("MSG ff:ff:ff:ff:ff:ff Dies ist eine Text-Nachricht!\n")

	## Callback for the [Send] button
	def cb_send(self, widget, data=None):
		(mac, name) = self.get_sel_peer()
		fcd = gtk.FileChooserDialog("Send file to %s..." % name, self.win,
				gtk.FILE_CHOOSER_ACTION_OPEN,
				(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
				 gtk.STOCK_OPEN, gtk.RESPONSE_OK))
		res = fcd.run()
		if res == gtk.RESPONSE_OK:
			if SEND_ABSOLUTE_PATH:
				fn = fcd.get_filename()
			else:
				fn = fcd.get_filename().split('/')[-1]
			self.send("SEND %s %s\n" % (mac, fn))
		fcd.destroy()
		#self.file(mac, name)

	## Main function calling the GTK event loop.
	def main(self):
		gtk.main()

if __name__ == "__main__":
	fe = Frontend()
	fe.main()

